#include	<P17CXX.H>
#include	"flash.h"
#include	"equates.h"
#include	"mydelay.h"
#include	"usart.h"

void SetupFlash(void)
{
	unsigned char dummy;

	//read RA3 to set it to an input
	#define FLASH_RY_BYBAR	PORTAbits.RA3
	#define FLASH_BUSY			!RA3
	dummy=FLASH_RY_BYBAR;
	
	//select correct bank by driving A16 low
	FLASH_A16_DIR=OUTPUT;											//make RB5 output
	FLASH_A16=0;															//set A16 on flash to zero
}

void EraseFlash(void)
{
	reset_flash;
	romuint(0xd555)=0xaaaa;
	romuint(0xaaaa)=0x5555;
	romuint(0xd555)=0x8080;
	romuint(0xd555)=0xaaaa;
	romuint(0xaaaa)=0x5555;
	romuint(0xd555)=0x1010;

	Delay10Us(60);
	wait_for_flash_finish;
	reset_flash;
	return;
}

unsigned int FlashDeviceID(void)
{
	unsigned int device_id;
	
	reset_flash;
	romuint(0xd555)=0xaaaa;
	romuint(0xaaaa)=0x5555;
	romuint(0xd555)=0x9090;
	return romuint(0xFF01);

	device_id=romuint(0xFF01);
	reset_flash;
	return device_id;
}

void api_flash_device_id(void)
{
	reset_flash;
	romuint(0xd555)=0xaaaa;
	romuint(0xaaaa)=0x5555;
	romuint(0xd555)=0x9090;
	api_flash_uint_data=romuint(0xFF01);
	reset_flash;
}

unsigned int FlashManufacturerID(void)
{
	unsigned int manufacturer_id;
	reset_flash;
	romuint(0xd555)=0xaaaa;
	romuint(0xaaaa)=0x5555;
	romuint(0xd555)=0x9090;
	manufacturer_id=romuint(0xFF00);
	reset_flash;
	return manufacturer_id;
}

void api_flash_manufacturer_id(void)
{
	reset_flash;
	romuint(0xd555)=0xaaaa;
	romuint(0xaaaa)=0x5555;
	romuint(0xd555)=0x9090;
	api_flash_uint_data=romuint(0xFF00);
	reset_flash;
}

//see AM29F400 data sheet for list of sectors

/*

Valid banks (anything else will overwrite program code)

Bank		Address range
4				08000h-0FFFFh
6				18000h-1FFFFh
8				28000h-2FFFFh
10			38000h-3FFFFh

*/

void EraseSector(unsigned int sector)
{
	reset_flash;
	romuint(0xf555)=0xaaaa;
	romuint(0xf2aa)=0x5555;
	romuint(0xf555)=0x8080;
	romuint(0xf555)=0xaaaa;
	romuint(0xf2aa)=0x5555;

	romuint((unsigned int)(sector))=0x3030;
	
	wait_for_flash_finish;
	reset_flash;
	return;
}

void api_erase_sector(void)
{
	reset_flash;
	romuint(0xf555)=0xaaaa;
	romuint(0xf2aa)=0x5555;
	romuint(0xf555)=0x8080;
	romuint(0xf555)=0xaaaa;
	romuint(0xf2aa)=0x5555;
	romuint(api_flash_uint_data)=0x3030;
	
	wait_for_flash_finish;
	reset_flash;
}

unsigned char WriteFlash(unsigned int addr, unsigned int data)
{
	unsigned char counter;
	unsigned char writeflash_uchar_timeout;

	//if routine tries to write to addresses which
	//  are in the PIC17Cx internal memory, the finished
	//  lines will never get toggled, resulting in an endless loop
	if (addr<FLASH_MIN_ADDRESS)
	{
		return FALSE;
	}
	
	counter=10;
	while(counter-->1)
	{
		//flash requires a preamble before each word programmed
		romuint(0xd555) = 0xaaaa;
		romuint(0xaaaa) = 0x5555;
		romuint(0xd555) = 0xa0a0;

		//put data on address lines
		romuint(addr) = data;
		
		//wait for data to be written (usually 15.6us)
		writeflash_uchar_timeout=0;
		#define WRITE_FLASH_TIMEOUT_MAX 0xF0		//@16Mhz, normally 3, F0 gives 1.26ms

		do
		{
			Nop();											//give flash time to realise that it's busy
			Nop();
			writeflash_uchar_timeout++;
		} while ( (flash_busy) && (writeflash_uchar_timeout <= WRITE_FLASH_TIMEOUT_MAX) );
		b;

		if (writeflash_uchar_timeout>=WRITE_FLASH_TIMEOUT_MAX)
		{
			return FALSE;
		}

		if(romuint(addr)==data) 												//make sure data was programmed
		{
			return TRUE;																	//if not try to reprogram
		}
	}
	return FALSE;
}

void api_write_flash(void)
{
	//if routine tries to write to addresses which
	//  are in the PIC17Cx internal memory, the finished
	//  lines will never get toggled, resulting in an endless loop

	if (api_flash_uint_address<FLASH_MIN_ADDRESS)
	{
		api_flash_uchar_return=FALSE;
		return;
	}
	
	api_flash_uchar_working=10;
	while(api_flash_uchar_working-->1)
	{
		//flash requires a preamble before each word programmed
		romuint(0xd555) = 0xaaaa;
		romuint(0xaaaa) = 0x5555;
		romuint(0xd555) = 0xa0a0;

		//put data on address lines
		romuint(api_flash_uint_address) = api_flash_uint_data;
		
		//wait for data to be written
		wait_for_flash_finish;

		if(romuint(api_flash_uint_address)==api_flash_uint_data)			//make sure data was programmed
		{
			api_flash_uchar_return=TRUE;
			return;																											//if not try to reprogram
		}
	}
	api_flash_uchar_return=FALSE;
	return;
}

//the unsigned int ReadFlash(unsigned int addr) function is simply #define uintrom(addr)

//reads a single byte in flash, see flsh_api.c
void api_read_flash(void)
{
	api_flash_uint_data=romuint(api_flash_uint_address);
	return;
}

//writes an array of unsigned ints from ram, &ram_addr to rom, rom_addr, of length length
unsigned char WriteFlashArray(unsigned int rom_addr, unsigned int *ram_addr, unsigned char length)
{
	unsigned char counter,retries;
	unsigned char writeflash_uchar_timeout;

	//if routine tries to write to addresses which
	//  are in the PIC17Cx internal memory, the finished
	//  lines will never get toggled, resulting in an endless loop
	if (rom_addr<FLASH_MIN_ADDRESS)
	{
		return FALSE;
	}
	
	counter=0;
	retries=0;								//a word is retried for programming
	while(counter<length)
	{
		//flash requires a preamble before each word programmed
		romuint(0xd555) = 0xaaaa;
		romuint(0xaaaa) = 0x5555;
		romuint(0xd555) = 0xa0a0;

		//put data on address lines
		romuint(rom_addr) = ram_addr[counter];
		
		//wait for data to be written (usually 15.6us)
		writeflash_uchar_timeout=0;
		#define WRITE_FLASH_TIMEOUT_MAX 0xF0		//@32Mhz, normally 3, F0 gives 0.63ms
		
		do
		{
			Nop();																//give flash time to realise that it's busy
			Nop();
			writeflash_uchar_timeout++;
		} while ( (flash_busy) && (writeflash_uchar_timeout <= WRITE_FLASH_TIMEOUT_MAX) );
		b;

		//if flash failed, return error as soon as possible
		if (writeflash_uchar_timeout>=WRITE_FLASH_TIMEOUT_MAX)
		{
			return FALSE;
		}

		if(romuint(rom_addr)==ram_addr[counter])				//make sure data was programmed
		{
			counter++;
			rom_addr++;
		}
		else
		{
			retries++;																		//if not try to reprogram, but only a limited no of times
			if (retries>10)
			{
				return FALSE;
			}
		}
	}
	return TRUE;
}

//writes an array of unsigned ints from ram, &ram_addr to rom, rom_addr, of length length
//rom_addr is passed in api_flash_uint_address, which becomes rom_pointer
//&ram_addr is passed in api_flash_uint_data, which becomes ram_pointer
//length is passed in api_flash_uchar_return (this saves ram) which becomes length
//api_flash_uint_lo_working, api_flash_uint_lo_working, api_flash_uchar_working are scratch pads

void api_write_flash_array(void)
{
	#define rom_pointer									api_flash_uint_address
	#define ram_pointer									api_flash_uint_data
	#define counter											api_flash_uint_hi_working
	#define retries 										api_flash_uint_lo_working
	#define	writeflash_uchar_timeout		api_flash_uchar_working			//api_flash_uchar_working is separate from the int_working
	#define	length											api_flash_uchar_return			//saves ram by using return value for length

	//if routine tries to write to addresses which
	//  are in the PIC17Cx internal memory, the finished
	//  lines will never get toggled, resulting in an endless loop
	if (rom_pointer<FLASH_MIN_ADDRESS)
	{
		api_flash_uchar_return=FALSE;
		return;
	}

	counter=0;
	retries=0;								//a word is retried for programming

	while(counter<length)
	{

		//flash requires a preamble before each word programmed
		romuint(0xd555) = 0xaaaa;
		romuint(0xaaaa) = 0x5555;
		romuint(0xd555) = 0xa0a0;

		//put data on address lines
		romuint(rom_pointer) = ramuint(ram_pointer);
		
		//wait for data to be written (usually 15.6us)
		writeflash_uchar_timeout=0;
		#define WRITE_FLASH_TIMEOUT_MAX 0xF0
		//@32Mhz, normally 3, F0 gives 0.63ms

		do
		{
			Nop();																//give flash time to realise that it's busy
			Nop();
			writeflash_uchar_timeout++;
		} while ( (flash_busy) && (writeflash_uchar_timeout <= WRITE_FLASH_TIMEOUT_MAX) );
		b;

		//if flash failed, return error as soon as possible
		if (writeflash_uchar_timeout>=WRITE_FLASH_TIMEOUT_MAX)
		{
			api_flash_uchar_return=FALSE;
			return;
		}

		if(romuint(rom_pointer)==ramuint(ram_pointer))	//make sure data was programmed
		{
			counter++;
			rom_pointer++;																//increment pointer to rom (external flash)
			ram_pointer+=2;																//increment pointer to ram twice, as it points to bytes
		}
		else
		{
			retries++;																		//if not try to reprogram, but only a limited no of times
			if (retries>10)
			{
				api_flash_uchar_return=FALSE;
				return;
			}
		}
	}

	api_flash_uchar_return=TRUE;
	return;

	//make sure the defines dont interfere with anything else
	//	as this would be a completely disastrous bug to fix
	#undef	rom_pointer
	#undef	ram_pointer
	#undef	counter
	#undef	retries
	#undef	writeflash_uchar_timeout
	#undef	length										//length is used elsewhere
}

//reads an array of unsigned ints from ram, &ram_addr to rom, rom_addr, of length length
void ReadFlashArray(unsigned int rom_addr, unsigned int *ram_addr, unsigned char length)
{
	unsigned char i;

	for(i=0;i<length;i++)
	{
		ram_addr[i]=ReadFlash(rom_addr);
		rom_addr++;
	}
}

//reads an array of unsigned ints from rom, rom_addr, to ram, &ram_addr, of length length
void api_read_flash_array(void)
{
	#define rom_pointer									api_flash_uint_address
	#define ram_pointer									api_flash_uint_data
	#define counter											api_flash_uint_hi_working
	#define	length											api_flash_uchar_return			//keeps it consistent with api_write_flash_array

	for(counter=0;counter<length;counter++)
	{
		ramuint(ram_pointer) = romuint(rom_pointer);
		ram_pointer+=2;		//ram is addressed in bytes, rom in words
		rom_pointer++;
	}

	//make sure the defines dont interfere with anything else
	//	as this would be a completely disastrous bug to fix
	#undef	rom_pointer
	#undef	ram_pointer
	#undef	counter
	#undef	length										//length is used elsewhere
}

//tests flash thoroughly, returns true on success
unsigned char TestFlashThorough(void)
{
	unsigned int i,j;
	unsigned int TestFlash_ViewData;
	unsigned int test_flash_thorough_pattern_offset;
	unsigned int test_flash_thorough_total_errors;

	#define TEST_FLASH_THOROUGH_TOTAL_ERRORS_MAX 50
	
	//choose random number for offset
	test_flash_thorough_pattern_offset=(TMR0H<<8)+TMR0L;
//	EraseFlash();

	test_flash_thorough_total_errors=0;
	ResetFlash();

	for (j=0;j<=1;j++)
	{
		SelectFlashBank(j);
	
		for(i=0x1000; i < 0xffff; i++)
		{
			if (WriteFlash(i,(i+test_flash_thorough_pattern_offset))==FALSE)
			{
				putch('!');
				putinthex(i);
				test_flash_thorough_total_errors++;
				if (test_flash_thorough_total_errors>TEST_FLASH_THOROUGH_TOTAL_ERRORS_MAX)
				{
					putch('-');
					putinthex(test_flash_thorough_pattern_offset);
					putch('-');
					putchhex(j);
					putch('-');
					return FALSE;
				}
			}
		}

		b;

		for(i=0x1000; i < 0xffff; i++)
		{
			TestFlash_ViewData=ReadFlash(i);
			if (TestFlash_ViewData != (i+test_flash_thorough_pattern_offset))
			{
				putch('@');
				putinthex(i);
				putch('#');
				putinthex((i+test_flash_thorough_pattern_offset));
				putch('$');
				putinthex(TestFlash_ViewData);
				test_flash_thorough_total_errors++;
				if (test_flash_thorough_total_errors>TEST_FLASH_THOROUGH_TOTAL_ERRORS_MAX)
				{
					putch('-');
					putinthex(test_flash_thorough_pattern_offset);
					putch('-');
					return FALSE;
				}
			}
		}
	}
	
	b;

	return TRUE;
}

unsigned char TestFlashBrief(void)
{
	unsigned int i;
	unsigned int TestFlash_ViewData;
	
	EraseFlash();
	
	for(i=0x8000; i < 0xf000; i+=0x901)
	{
		WriteFlash(i,(i+0x4000));
	}

	b;

	for(i=0x8000; i < 0xf000; i+=0x901)
	{
		TestFlash_ViewData=ReadFlash(i);
		if (TestFlash_ViewData != (i+0x4000))
		{
			return FALSE;
		}
	}

	b;
	return TRUE;
}

/*********************************************************************/
/* Flash_status utilizes the DQ6, DQ5, and DQ3 polling algorithms    */
/* described in the flash data book.  It can quickly ascertain the   */
/* operational status of the flash device, and return an             */
/* appropriate status code (defined in flash.h)                      */
/*********************************************************************/
 
unsigned char FlashStatus(void)
{
	unsigned int d, t;
	unsigned int retry;

	retry=0xffff;

	//retry if there is a write completion
	while(retry>1)
	{
		d = romuint(0xaaaa);        				/* read data */
		t = d ^ romuint(0xaaaa);    				/* read it again and see what toggled */

		if (t == 0) {           						/* no toggles, nothing's happening */
			return STATUS_READY;
		}
		else if (t == 0x04) { 							/* erase-suspend */
			if (retry<5) 
			{		  														/* may have been write completion */
					return STATUS_ERSUSP;
			}
		}
		else if (t & 0x40) {
			if (d & 0x20) {     							/* timeout */
				return STATUS_TIMEOUT;
			}
			else {
				return STATUS_BUSY;
			}
			retry--;
		}
	}
	return STATUS_ERROR;
}
