/*

MaiN PiC designed for Hitech c (MNPICHT)

Processor:				pic17c44
Compiler:					hi-tech C v7.83pl3
Environment:			compiled with MPLab 4.00.14
Description:			this project is an example of how to generate a .hex file for loading
									into external AM29F400 flash
									there is examples of how to write to flash utilising calling routines
									in the original boot code (see BTPICHT project)
Compiler options:	-G (generate debug), global optimizaitons (enter 9 in data column), local optimizations
Linker options:		-L-Pmainhook=03FFEh -L-Pidword=03FFCh -L-ACODE=01000h-01FFCh  -L-AROMDATA=01000h-01FFCh
Optimisation:			this code is not optimised for speed, it needs some routines converted to
									#defines and the total drop path of procedures shortened
Author:						Shane Tolmie
Tested:						as part of system to download and run code from flash for remote web updates
									data acquisition portion tested OK
									
*/

#define BUILD_VERSION 3

//put the address of main() at the very end of the eeprom space
//since a pic17c44 has 8k of rom, this is word 0x1FFF, byte 0x3FFE
//or location 0x1FFE in the .hex file
//the boot loader uses this to find the address to jump to for main
//this must be accompanied by the option -L-Pmainhook=03FFEh in the linker options
asm("psect mainhook,delta=1,size=2");
asm("main_hook DW _main");

//put an idenfifying word at the end of the hex file, to check if all is well
//this has the added advantage of making sure that the main_hook is correctly
//set in the linker functions
//this must be accompanied by the option -L-Pidword=03FFCh in the linker options
asm("psect idword,delta=1,size=2");
asm("id_word DW 0xdea1");

#include <pic.h>

//*************
//include files

#include	"codes.h"
#include	"equates.h"
#include	"serial.h"
#include	"flash_bh.h"
#include	"always.h"
#include	"string.h"

unsigned char delayus_variable;
#include	"delay.h"

// Preprocessor macro definitions

//global variables

unsigned int i;
//extern unsigned int write_flash_data;


//************
//main program

void main(void)
{
	init_serial();

	putlf;putlf;
	putstlf("Main PIC operating off boot routines, Compiled with HI-TECH C v7.83pl3");
	putst("Build ");putchdec(BUILD_VERSION);putstlf(", (c)1999 Shane Tolmie");
	putlf;

	putstlf("Testing flash routines ...");
	putlf;putlf;

	API_TestRoutines();
	while(1);
}

